/** \file votable.h
 * \brief libVOTable header file
 */
//-------------------------------------------------------------
//- Project: libVotable
//- Filename: votable.c
//-------------------------------------------------------------
//- Author: J-C MALAPERT
//- Creation date: 25/11/2004
//- Last modification date: 02/12/2004
//-------------------------------------------------------------
//- Comments:
//- 25/11/2004    Creation
//- 02/12/2004    - fix bug about memory free
//                - init list in Init_VO_Parser
//                - memory allocation for column in 
//                  Extract_VO_Header
//-------------------------------------------------------------

#ifndef _VOTABLE_H_
#define _VOTABLE_H_
#include <stdio.h>
#include <stdlib.h>
#include <libxml/xmlmemory.h>
#include <libxml/xmlreader.h>

/* --------------------------------- Macros ---------------------------- */

#define DEBUG(x) printf(x)
/** \def RETURN_OK
 *  \brief One of Codes returned by functions
 *
 * is returned when the function successed
 */
#define RETURN_OK 0

/** \def RETURN_ERROR
 *  \brief One of Codes returned by functions
 *
 * is returned when the function failed
 */
#define RETURN_ERROR 1

/** \def EXIT_MEMORY
 *  \brief One of Codes returned by functions
 * 
 * is return when there is not enough memory
 */
#define EXIT_MEMORY 2

/** \def EXIT_READING
 *  \brief Can't open file
 *
 *  exit with EXIT_MEMORY value when xmlReaderForFile function failed
 */
#define EXIT_READING 3

/** \def QMALLOC
 *  \brief Macro for memory allocation
 *
 */
#define QMALLOC(ptr, typ, nel) \
                {if (!(ptr = (typ *)malloc((size_t)(nel)*sizeof(typ)))) \
                  error(EXIT_MEMORY, "Not enough memory for ", \
                        #ptr " (" #nel " elements) !");;}



/* ------------------------- Structure definition -----------------------*/

/** \struct _list_field
 * \brief Structure dedicated to FIELD tag
 * 
 * Structure containing all attributes allowed for FIELD tag.\n 
 * In order to save all this structure type, we create a linking list of this one.
 */

typedef struct _list_field list_field;

struct _list_field {
  xmlChar *ID;        /**< \brief  ID used : implied */
  xmlChar *unit;      /**< \brief  unit used : implied */
  xmlChar *datatype;  /**< \brief  datatype used  : implied
		       *\n
		       * among  : boolean, bit, unsignedByte, short ,int, long, char
		       * unicodeChar, float, double, floatComplex or doubleComplex 
		       */
  xmlChar *precision; /**< \brief  value precision used : implied */
  xmlChar *width;     /**< \brief  value width used : implied */
  xmlChar *ref;       /**< \brief  ref attribute used : implied */
  xmlChar *name;      /**< \brief  name attribute used : implied */
  xmlChar *ucd;       /**< \brief  unified content type attribute used : implied */
  xmlChar *arraysize; /**< \brief  arraysize used : implied */
  xmlChar *type;      /**< \brief  type used among : implied 
		       *\n
		       * among : hidden, no_query or trigger */
  int position;      /**< \brief position 
		      *\n
		      * position of the FIELD element in respect to others ones
		      */
  list_field *next; /**< \brief address of the next element of the list */
};


/** \struct _list_tabledata
 * \brief Structure dedicated to TABLEDATA tag 
 * 
 * The TABLEDATA does not contain attributes. Moreover, only the tag
 * embedded in the TABLEDATA element contains a PCDATA value and
 * attribute. For this reason and to reduce the number of structures, the
 * TD structure is the TABLEDATA structure containing all attributes
 * allowed for TD tag and its associated value.  In order to save all this
 * structure type, we create a linking list.  
 */
typedef struct _list_tabledata list_tabledata;

struct _list_tabledata {
  xmlChar *value;     /**< \brief TD tag value used*/
  xmlChar *ref;       /**< \brief TABLEDATA attribute used */
  int colomn;         /**< \brief column to parse */
  list_tabledata *next; /**< \brief address of the next element of the list */
};


/** \struct _list_table
 * \brief Structure dedicated to TABLE tag
 * 
 * Structure containing all attributes allowed for TABLE tag.\n 
 * In order to save all this structure type, we create a linking list of this one.
 */
typedef struct _list_table list_table;

struct _list_table {
  xmlChar *ID;         /**< \brief ID attribute used : implied */
  xmlChar *name;       /**< \brief name attribute used : implied */
  xmlChar *ref;        /**< \brief ref attribute used : implied */
};


/** \struct _VOTable
 * \brief Main structure
 *
 * Structure containing all structures relative to VOTABLE 
 */
typedef struct _VOTable VOTable;

struct _VOTable {
  list_table *table; /**< \brief structure dedicated to TABLE tag */
  list_tabledata *tabledata; /**< \brief structure dedicated to TABLEDATA tag */
  list_field *field; /**< \brief structure dedicated to FIELD tag */
};

/* -------------------------- Protos --------------------------*/

/*
  PURPOSE : insert attributes of FIELD elements in a list
  INPUT   : Pointer on reader | pointer on header | FIELD positon in respect to the others ones
  OUTPUT  : list
  AUTHORS : J-C Malapert
  VERSION : 25/11/2004
*/
static list_field *insert_field(xmlTextReaderPtr reader, 
				  list_field *list, 
				  int position);


/*
  PURPOSE : insert values of TABLEDATA in a list for a particular FIELD
  INPUT   : Pointer on reader | pointer on tabledata FIELD positon in respect to the others ones
  OUTPUT  : list
  AUTHORS : J-C Malapert
  VERSION : 25/11/2004
*/
static list_tabledata *insert_tabledata(xmlTextReaderPtr reader, 
					list_tabledata *list, 
					int position);


/*
  PURPOSE : insert values of TABLE attributes in a structure
  INPUT   : Pointer on reader
  OUTPUT  : list
  AUTHORS : J-C Malapert
  VERSION : 25/11/2004
*/
static list_table *insert_table(xmlTextReaderPtr reader);



/** \fn  static void Free_VO_Fields(list_field *vlist_field,int **column)
 * \brief Free field linking list and column\n
 * PURPOSE : Free field linking list and column
 *
 * \param vlist_field Pointer on field linking list 
 * \param column Array representing the columns position to extract in TABLEDATA party
 */ 
static void Free_VO_Fields(list_field *vlist_field, 
			   int **column);



/** \fn static void Free_VO_Tabledata(list_tabledata *vlist_tabledata)
 * \brief Free tabledata linking list\n
 * PURPOSE : Free tabledata linking list
 *
 * \param vlist_tabledata Pointer on field linking list 
 */ 
static void Free_VO_Tabledata(list_tabledata *vlist_tabledata);



/** \fn  sattic void Free_VO_Table(list_table *vlist_table)
 * \brief Free table linking list\n
 * PURPOSE : Free table linking list
 *
 * \param vlist_table Pointer on field linking list 
 */ 
static void Free_VO_Table(list_table *vlist_table);


/** @defgroup libintroduction Introduction to the libVOTable library
 *   The VOTable Library consists in a \b single \b header \b file \ref votable.h providing a set of C functions that
  can be used in your own sources, to manage memory, parse document. It is portable 
  , efficient and  simple to use.

  \section s1 Library structure

  The file \ref votable.h contains all the functions that compose the library itself. It is organized as follows :
  - Define indicating the value returned by the library. 
  \code
  #define RETURN_OK 0
  #define RETURN_ERROR 1
  #define EXIT_MEMORY 2
  #define EXIT_READING 3
  \endcode
  - All functions for memory managing
  - All functions to extract attributes and values in VOTable format

  To use this library in your own C code, include the following line :
  \code
  #include "votable.h"
  \endcode
  

  \section s2 How to parse a VOTable file with the libVOTable

  To parse a VOTable file with the libVOTable, follow the instructions :
  - First, initialise memory and structures with \ref Init_VO_Parser function. 
  - Then use \ref Extract_VO_Fields to load in memory all attributes of each \b FIELD elements. Browse memory with a pointer and write the \b returned FIELD position in a array. This array contains all data positions that you want to extract in \b TABLEDATA element.
  - Once it is filled, the array can be provided as a parameter to \ref Extract_VO_TableData function, that it will return the data back
  - If you want go on to parse (if it is possible) or jump a part of the file, you can do so the \ref Move_to_Next_VO_Fields and  \ref Move_to_Next_VO_Table functions
  - When you are at the end with the VOTable parser, don't forget to free memory ;-)

  To easier the functions usage, they have been grouped together into modules in the documentation. These modules are :
  - Initialization and memory management
  - Extract Attributes and values from a VOTABLE element
  - Jump to a VOTABLE tag

  
  \section s3 How to compile ?

  Thanks to its packaging, the libVOTable library is a very light and user-friendly library .
  The easiest way to compile your own code \b test.c with the libVOTable library is to type :\n
  \code
  libtool gcc -o test -static test.c libvotable.la -I/usr/local/include/libxml2 -lxml2
  \endcode

  \section s4 What's next ?

  If you are ready to get more, and to start writing more serious programs
  with libVOTable, you are invited to go to the \ref libvotable_tutorial section.

**/
/*@}*/

/** @defgroup libvotable_tutorial Tutorial : Getting Started
 *   Assume you need to read the following VOTable file :\n
\code
<?xml version="1.0"?>
<VOTABLE version="1.1" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
 xsi:noNamespaceSchemaLocation="http://www.ivoa.net/xml/VOTable/VOTable/v1.1">
  <DESCRIPTION>Data to import to skywatcher</DESCRIPTION>
  <RESOURCE>
    <PARAM name="Instrument" datatype="char" arraysize="*" value="TOTO">
      <DESCRIPTION>
        This parameter is designed to store intrument's name
      </DESCRIPTION>
    </PARAM>
    <TABLE name="SpectroLog">
      <FIELD name="Target" ucd="meta.id" datatype="char" arraysize="30*"/>
      <FIELD name="Instr" ucd="instr.setup" datatype="char" arraysize="5*"/>
      <FIELD name="Dur" ucd="time.expo" datatype="int" width="5" unit="s"/>
      <FIELD name="Spectrum" ucd="meta.ref.url" datatype="float" arraysize="*"
             unit="mW/m2/nm" type="location">
        <DESCRIPTION>Spectrum absolutely calibrated</DESCRIPTION>
        <LINK type="location"
              href="http://ivoa.spectr/server?obsno="/>
      </FIELD>
      <DATA><TABLEDATA>
        <TR><TD>NGC6543</TD><TD>SWS06</TD><TD>2028</TD><TD>01301903</TD></TR>
        <TR><TD>NGC6543</TD><TD>SWS07</TD><TD>2544</TD><TD>01302004</TD></TR>
      </TABLEDATA></DATA>
    </TABLE>
  </RESOURCE>
</VOTABLE>
\endcode

  The code below that, uses the libVOTable library, does the task :
\code
#include "votable.h"

int main() {
  xmlTextReaderPtr reader;
  list_field *vfield_move;
  list_tabledata *vtabledata_move;
  VOTable votable;
  int nbFields;
  int *columns;
  char file[50]="votable.xml";
  reader = Init_VO_Parser(file,&votable);

  Extract_Att_VO_Table(reader,&votable);
  printf("Table Attribute=%s\n\n",votable.table->name);

  Extract_VO_Fields(reader,&votable,&nbFields,&columns);
  for(vfield_move=votable.field;vfield_move!=NULL;vfield_move=vfield_move->next) {
    printf("name=%s\nucd=%s\ndatatype=%s\narraysize=%s\ntype=%s\nwidth=%s\nunit=%s\n\n",
	   vfield_move->name,
	   vfield_move->ucd,
	   vfield_move->datatype,
	   vfield_move->arraysize,
	   vfield_move->type,
	   vfield_move->width,
	   vfield_move->unit);
    if(xmlStrcmp(vfield_move->ucd,"meta.id") == 0) 
      columns[0] = vfield_move->position;
    if(xmlStrcmp(vfield_move->ucd,"meta.ref.url") == 0) 
      columns[1] = vfield_move->position;
 }


 Extract_VO_TableData(reader,&votable, nbFields, columns);
 for(vtabledata_move=votable.tabledata;vtabledata_move!=NULL;vtabledata_move=vtabledata_move->next) {  
   printf("All values=%s\n",vtabledata_move->value);
   if (vtabledata_move->colomn == columns[0])
     printf("ucd=meta.id value=%s\n",vtabledata_move->value);
   if (vtabledata_move->colomn == columns[1])
     printf("ucd=meta.ref.url value=%s\n",vtabledata_move->value);
 }

  if (Free_VO_Parser(reader,&votable,&columns) == 1)
    fprintf(stderr,"memory problem\n");
  return 0;
}
\endcode

The program output is :
\code
Table Attribute=SpectroLog

name=Spectrum
ucd=meta.ref.url
datatype=float
arraysize=*
type=location
width=(null)
unit=mW/m2/nm

name=Dur
ucd=time.expo
datatype=int
arraysize=(null)
type=(null)
width=5
unit=s

name=Instr
ucd=instr.setup
datatype=char
arraysize=5*
type=(null)
width=(null)
unit=(null)

name=Target
ucd=meta.id
datatype=char
arraysize=30*
type=(null)
width=(null)
unit=(null)

All values=01302004
ucd=meta.ref.url value=01302004
All values=NGC6543
ucd=meta.id value=NGC6543
All values=01301903
ucd=meta.ref.url value=01301903
All values=NGC6543
ucd=meta.id value=NGC6543
\endcode

Detailled description of each task, line by line :
\code
#include "votable.h"
\endcode
Include the main and only header file of the libVOTable library.
\code 
int main(){
\endcode
Definition of the main function.
\code
xmlTextReaderPtr reader;
\endcode
the xmlTextReaderPtr used
\code
list_field *vfield_move;
\endcode
Pointer on FIELD datatype
\code
list_tabledata *vtabledata_move;
\endcode
Pointer on TABLEDATA datatype
\code
VOTable votable;
\endcode
VOTABLE structure declaration
\code
int nbFields;
\endcode
Number of FIELD
\code
int *columns;
\endcode
Integer array of TD column number to parse
\code
char file[50]="votable.xml";
\endcode
Filename to parse
\code
reader = Init_VO_Parser(file,&votable);
\endcode
Memory and structure initialization
\code
Extract_Att_VO_Table(reader,&votable);
\endcode
Extract TABLE attributes
\code
printf("Table Attribute=%s\n\n",votable.table->name);
\endcode
Display result
\code
Extract_VO_Fields(reader,&votable,&nbFields,&columns);
\endcode
Extract FIELD attributes
\code
for(vfield_move=votable.field;vfield_move!=NULL;vfield_move=vfield_move->next) {
  printf("name=%s\nucd=%s\ndatatype=%s\narraysize=%s\ntype=%s\nwidth=%s\nunit=%s\n\n",
          vfield_move->name,
          vfield_move->ucd,
          vfield_move->datatype,
          vfield_move->arraysize,
          vfield_move->type,
          vfield_move->width,
          vfield_move->unit);
  if(xmlStrcmp(vfield_move->ucd,"meta.id") == 0) 
      columns[0] = vfield_move->position;
  if(xmlStrcmp(vfield_move->ucd,"meta.ref.url") == 0) 
      columns[1] = vfield_move->position;
}
\endcode
Browse the FIELD linking list and prepare the parsing for data for which ucd are meta.id and meta.ref.url 
\code
Extract_VO_TableData(reader,&votable, nbFields, columns);
\endcode
Extract data for which <a href="http://vizier.u-strasbg.fr/w/doc/UCD/">ucd</a> are meta.id and meta.ref.url
\code
for(vtabledata_move=votable.tabledata;vtabledata_move!=NULL;vtabledata_move=vtabledata_move->next) {  
  printf("All values=%s\n",vtabledata_move->value);
  if (vtabledata_move->colomn == columns[0])
    printf("ucd=meta.id value=%s\n",vtabledata_move->value);
  if (vtabledata_move->colomn == columns[1])
    printf("ucd=meta.ref.url value=%s\n",vtabledata_move->value);
}
\endcode
Browse the TABLEDATA linking list and display parsing result
\code
if (Free_VO_Parser(reader,&votable,&columns) == 1)
   fprintf(stderr,"memory problem\n");
\endcode
Free parser memory
**/
/*@}*/

/** @defgroup libvomemory Initialization and memory management
 * Parser initializing and free memory
*/

/*@{*/
/** \fn xmlTextReaderPtr Init_VO_Parser(const char *filename,VOTable *votablePtr)
 * \brief Parser initialization\n
 * PURPOSE : Initializing of VOTable structure and creating xmlTextReader pointer
 *
 * \param filename VOTABLE filename
 * \param votablePtr Pointer on VOTable structure
 *
 * \exception EXIT_READING can't read filename and exit program
 * 
 * \return xmlTextReaderPtr Pointer on xmlTextReader\n
 * _VOTable structure initialized
 */ 
xmlTextReaderPtr Init_VO_Parser(const char *filename,
				VOTable *votablePtr);



/** \fn int  Free_VO_Parser(xmlTextReaderPtr reader,VOTable *votablePtr,int **column)
 * \brief VO_Parser memory free\n
 * PURPOSE : Free VOTable structure, column and xmlTextReader
 *
 * \param reader Pointer on xmlTextReader
 * \param votablePtr Pointer on VOTable structure
 * \param column Array representing the columns position to extract embedded in TABLEDATA 
 *
 * \return RETURN_OK : free successed\n
 * RETURN_ERROR : xmlTextReaderClose failed
 */ 
int Free_VO_Parser(xmlTextReaderPtr reader,
		   VOTable *votablePtr,
		   int **column);
/*@}*/




/** \defgroup libvoextract Extract Attributes and values from a VOTABLE element
 * Free memory if needed, extract data from VOTABLE file and store them in the _VOTable structure 
 */
/*@{*/
/** \fn void Extract_VO_Fields (xmlTextReaderPtr reader,VOTable *votablePtr,int *nbFields,int **columns)
 * \brief Procedure to extract FIELD attribute\n
 * PURPOSE : Free field linking list if needed, extracts FIELD attributes 
 * from VOTABLE file and stores them in VOTable structure
 *
 * \param reader Pointer on xmlTextReader\n
 * \param votablePtr Pointer on VOTable structure
 * \param nbFields FIELD tag occurence embedded in one TABLE element
 * \param columns Array representing the columns position to extract in TABLEDATA
 *
 * \return _VOTable structure filled\n
 * Field tag occurence in one TABLE tag\n
 * Array allocated to nbFields size and zero initialized 
 */ 
void Extract_VO_Fields ( xmlTextReaderPtr reader,
			 VOTable *votablePtr, 
			 int *nbFields, 
			 int **columns);



/** \fn void Extract_VO_TableData (xmlTextReaderPtr reader,VOTable *votablePtr,int nbcolumns,int *columns)
 * \brief Procedure to extract TD tag attributes and values\n
 * PURPOSE : Free tabledata linking list if needed, Extracts TD attributes and values 
 * from VOTABLE file and stores them in VOTable structure
 *
 * \param reader Pointer on xmlTextReader
 * \param votablePtr Pointer on VOTable structure
 * \param nbcolumns Field tag occurence embedded in one TABLE element
 * \param columns Array representing the columns position to extract in TABLEDATA tag
 *
 * \return _VOTable structure filled
 * \return Number of records
 */ 
int Extract_VO_TableData (xmlTextReaderPtr reader, 
			  VOTable *votablePtr,  
			  int nbcolumns, 
			  int *columns);



/** \fn void Extract_Att_VO_Table (xmlTextReaderPtr reader,VOTable *votablePtr)
 * \brief Procedure extracting TABLE tag attributes\n
 * PURPOSE : Free table linking list if needed, extracts TABLE attributes from VOTABLE file and stores them in VOTable structure
 *
 * \param reader Pointer on xmlTextReader
 * \param votablePtr Pointer on VOTable structure
 *
 * \return _VOTable structure filled
 * \return Number of records
 */ 
void Extract_Att_VO_Table(xmlTextReaderPtr reader, 
			  VOTable *votablePtr);
/*@}*/




/** \defgroup libvojump Jump to a VOTABLE tag
 * Functions to jump to a VOTABLE tag
 */
/*@{*/
/**\fn void Move_to_Next_VO_Fields (xmlTextReaderPtr reader)
 * \brief Move reader to the first FIELD tag contained in the next TABLE tag\n
 * PURPOSE : Try to move to the first FIELD tag embedded in the next TABLE tag
 *
 * \param reader Pointer on xmlTextReader
 *
 * \exception RETURN_ERROR can't move to the next TABLE tag\n
 * Move reader to the end of file
 * \return RETURN_OK Move reader to the first FIELD tag of the next TABLE tag
 */ 
int Move_to_Next_VO_Fields (xmlTextReaderPtr reader);



/** \fn  int Move_to_Next_VO_Table (xmlTextReaderPtr reader)
 * \brief Try to move to the next TABLE tag\n
 * PURPOSE : Try to move to the next TABLE tag
 *
 * \param reader Pointer on xmlTextReader
 *
 * \exception RETURN_ERROR can't move to the first FIELD embedded in the next TABLE element\n
 * Move reader to the end of file
 * \return RETURN_OK Move reader to the next TABLE tag 
 */ 
int Move_to_Next_VO_Table (xmlTextReaderPtr reader);
/*@}*/



#endif

/*! \mainpage
 
@section abstract Abstract

The libVOTable library primary task is to quickly parse big VOTABLE files that use a large amount of memory, for example, catalogues produced by big surveys.
The libVOTable library has functionalities to handle VOTABLE files associated to big data bulks. This is an optimised high-speed library, especially designed for memory saving. See the  <a href="introframe.html">Introduction</a> fore more details.
This TERAPIX tool is available to the community under the GPL licence as libVOTable.

@section dev Development

libVOTable is not yet finished. The following "to do list"
indicates we would like to implement

 \todo	  
  Parse all elements and attributes given by VOTABLE
 \n
  Check VOTABLE validity
 \n
  Implementing other parsing methods


@section Documentation

   
   This is the reference documentation of the libVOTable Library</a>.
   This documentation has been automatically generated from the header file votable.h, using the tool
   <a href="http://www.doxygen.org">doxygen</a>.
   It contains a detailed description of all functions of the libVOTable Library</a>.
   
   You can easily navigate through the documentation pages, using the menu above.

   To get started, you may first check the list of <a href="modules.html">available modules</a>.

@section contact Contact

libVOTable was developed by Jean-Christophe Malapert <a href="mailto:malapert@iap.fr?subject=libVOTable">malapert@iap.fr</a>

*/
